/*!
	\file
	FwSocket2.cpp : implementation file.
	$Author: Arcady Pchelintsev $
	$Date: 2001/10/19 11:16:54  $
	$State: Exp $
	$Source: \\PRAGMA\DATAC\FlexTelnet2\FTN\FwSocket2.cpp $
	$Revision:$	1.1.0.0
\n  $Log: $
*/

#include "stdafx.h"
#include "FwSocket2.h"
#include <process.h>

#pragma comment(lib,"ws2_32")

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define BUFFERSIZE	0x7FFF

static LONG s_dwConcurrentConnections=0;

///////////////////////////////////////////////////
// CBuffer implementation


IMPLEMENT_DYNCREATE(CFwSocket2,CObject)

 
CFwSocket2::CBuffer::CBuffer()
{
	m_uBufLen=0;
	m_uBufLenMax=BUFFERSIZE;
	m_pIp=NULL;
	m_pBuf=(BYTE *)realloc(NULL,m_uBufLenMax+1);
	m_pBuf[0]=0;
	InitializeCriticalSection(&m_Crs);
}

// CTcpIpBuffer denstructor
CFwSocket2::CBuffer::~CBuffer()
{
	DeleteCriticalSection(&m_Crs);
	free(m_pBuf);
}

// Buffer realocation
BOOL CFwSocket2::CBuffer::Realloc(UINT uSize)
{
	LPCSTR fname="Realloc(UINT)";
	BOOL retv=TRUE;

	if (uSize>=0x1000000) {
		m_pIp->OnError("Buffer size exeeded 1 Mb in CFwSocket::CBuffer::Realloc()");
		return FALSE;
	}

	Lock();
	BYTE *tmp=(BYTE *)realloc(m_pBuf,uSize);
	if (tmp) {
		m_pBuf=tmp;
		m_uBufLenMax=uSize;
	} else {
		m_pIp->OnError("Memory Allocation Error in CFwSocket::CBuffer::Realloc()");
		retv=FALSE;
	}

	Unlock();
	return retv;
}

// Add byte to buffer
BOOL CFwSocket2::CBuffer::Add(char ch)
{
	// add byte to buffer
	Lock();
	m_pBuf[m_uBufLen++]=ch;
	Unlock();

	return TRUE;
}

// Add string to buffer
BOOL CFwSocket2::CBuffer::Add(LPCSTR szStr)
{
	return Add((void *)szStr,strlen(szStr));
}

// Add binary data to buffer
BOOL CFwSocket2::CBuffer::Add(LPCVOID pvMem, UINT uLen)
{
	// add binary data to buffer
	if (m_uBufLen+uLen+1>m_uBufLenMax) 
		while (!Realloc(m_uBufLen+uLen+1)) {
			Sleep(20);
		}
	Lock();
	memcpy(&m_pBuf[m_uBufLen],pvMem,uLen);
	m_uBufLen+=uLen; m_pBuf[m_uBufLen]=0;
	Unlock();

	return TRUE;
}

// delete n bytes from buffer
BOOL CFwSocket2::CBuffer::Del(UINT uNumber,UINT uBegin)
{
	if (uNumber+uBegin>m_uBufLen) uNumber=m_uBufLen;
	Lock();
	for (UINT i=uBegin; i<m_uBufLen-uNumber; i++) m_pBuf[i]=m_pBuf[i+uNumber];
	m_uBufLen-=uNumber; m_pBuf[m_uBufLen]=0;
	Unlock();
	return TRUE;
}

// Purge any data from buffer ans realloc memory
void CFwSocket2::CBuffer::Purge()
{
	Lock();
	m_uBufLen=0;
	Unlock();
	Realloc(BUFFERSIZE);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

class CFwSocket2Init
{
public:
	CFwSocket2Init();
	virtual ~CFwSocket2Init();

	WSAPROTOCOL_INFO *GetInfo() {return m_pPROTOCOL_INFO;}
	int GetInfoCount() {return m_iInfoCount;}
	int GetFoundIndex() {return m_iFoundIndex;}
	BOOL IsQosFound() {return m_bQosProtocolFound;}
	WSADATA& GetWSADATA() {return m_WSADATA;}
protected:
	WSADATA m_WSADATA;						
	int m_iFoundIndex;						// found protocol index
	int m_iInfoCount;						// number of entries
	BOOL m_bQosProtocolFound;				// QOS protocol found
	WSAPROTOCOL_INFO *m_pPROTOCOL_INFO;		// protocol info buffer

};

CFwSocket2Init::CFwSocket2Init()
{
	m_bQosProtocolFound=FALSE;
	m_pPROTOCOL_INFO=NULL;
	m_iInfoCount=0;
	m_iFoundIndex=-1;

	WSADATA m_WSADATA;		// for WSA init
	if (WSAStartup(MAKEWORD(2,0),&m_WSADATA)) return;

#ifdef QUALITY_SERVICE

	DWORD dwBufLen=0;
	DWORD dwErr;

	// First, have WSAEnumProtocols tell you how big a buffer you need.
	int iRet=WSAEnumProtocols(NULL, m_pPROTOCOL_INFO, &dwBufLen);
	if (SOCKET_ERROR != iRet) {
         TRACE("WSAEnumProtocols: should not have succeeded\n");
	} else {
		if (WSAENOBUFS != (dwErr = WSAGetLastError())) {
			// WSAEnumProtocols failed for some reason not relating to buffer
			// size - also odd.
			TRACE("WSAEnumProtocols(1): %d\n", WSAGetLastError());
		} else {
			// WSAEnumProtocols failed for the "expected" reason. Therefore,
			// you need to allocate a buffer of the appropriate size.
			m_pPROTOCOL_INFO=(WSAPROTOCOL_INFO *)new char[dwBufLen];
			if (m_pPROTOCOL_INFO) {
				// Now you can call WSAEnumProtocols again with the expectation
				// that it will succeed because you have allocated a big enough
				// buffer.
				iRet = WSAEnumProtocols(NULL, m_pPROTOCOL_INFO, &dwBufLen);
				if (SOCKET_ERROR == iRet) {
					TRACE("WSAEnumProtocols(3): %d\n", WSAGetLastError());
				}
				else {
					// Loop through protocols, looking for the first service
					// provider that meets the matching criteria.
					m_iInfoCount=iRet;
					for (int i=0; i<iRet; i++) {
						TRACE("Protocol %1: '%s'\n",m_pPROTOCOL_INFO[i].szProtocol);
						if ((IPPROTO_TCP == m_pPROTOCOL_INFO[i].iProtocol)) {
							m_iFoundIndex=i;
							if ((XP1_QOS_SUPPORTED==(XP1_QOS_SUPPORTED & m_pPROTOCOL_INFO[i].dwServiceFlags1))) {
								m_bQosProtocolFound=TRUE;
								break;
							}
						}
					}
				}
			}
		}
	}

#endif

}

CFwSocket2Init::~CFwSocket2Init()
{
	if (m_pPROTOCOL_INFO) {
		delete[] m_pPROTOCOL_INFO;
	}

	WSACleanup();
}

static CFwSocket2Init s_FwSocket2Init;		// static object to inicialize Windows Sockets


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CFwSocket2::CFwSocket2()
{
	InitializeCriticalSection(&m_CrsPut);

	m_liMaxRecvSilence=WSA_INFINITE;
	m_dwMaxConnections=20;
	m_bConnected=FALSE;
	m_bAcceptedSock=FALSE;
	m_hShellProcess=INVALID_HANDLE_VALUE;
	m_hSocket=INVALID_SOCKET;
	m_hEvent=WSACreateEvent();
	m_hEventIoThread=WSACreateEvent();
	m_hEventDataReady=WSACreateEvent();
	m_hEventStopThread=WSACreateEvent();
	ZeroMemory(&m_WSAOVERLAPPED,sizeof(m_WSAOVERLAPPED));
	m_WSAOVERLAPPED.hEvent=WSACreateEvent();
	m_hThread=INVALID_HANDLE_VALUE;
	ZeroMemory(&m_Stats,sizeof(m_Stats));

	ZeroMemory(&m_AddrAccepted,sizeof(m_AddrAccepted));
	m_AddrAccepted.sa_family=AF_INET;
	m_InBuf.m_pIp=this;
}

CFwSocket2::~CFwSocket2()
{
	Disconnect();

	WSACloseEvent(m_hEvent);
	WSACloseEvent(m_hEventIoThread);
	WSACloseEvent(m_hEventStopThread);
	WSACloseEvent(m_WSAOVERLAPPED.hEvent);
	WSACloseEvent(m_hEventDataReady);
	DeleteCriticalSection(&m_CrsPut);
}

void CFwSocket2::InitSocket(BOOL bServerSocket)
{
	int iIndex=s_FwSocket2Init.GetFoundIndex();
	WSAPROTOCOL_INFO *pInfo=s_FwSocket2Init.GetInfo();

	if (s_FwSocket2Init.IsQosFound()) {
		m_hSocket=WSASocket(FROM_PROTOCOL_INFO,FROM_PROTOCOL_INFO,
			FROM_PROTOCOL_INFO,&pInfo[iIndex],0,WSA_FLAG_OVERLAPPED);
	} else {
		m_hSocket=WSASocket(AF_INET,SOCK_STREAM,IPPROTO_TCP,NULL,0,0);
	}

    if (INVALID_SOCKET==m_hSocket) {
		OnError("Error WSASocket()");
		return;
	}

	// switching socket to non-blocking mode
	ULONG uOne=1,uTmp=1;
	WSAIoctl(m_hSocket,FIONBIO,&uOne,sizeof(uOne),&uTmp,sizeof(uTmp),&uTmp,NULL,NULL);

	// switching socket to keepalive mode
	uOne=1;
	setsockopt(m_hSocket,SOL_SOCKET,SO_KEEPALIVE,(char *)&uOne, sizeof(uOne));

	if (bServerSocket) {
		uOne=1;
		setsockopt(m_hSocket,SOL_SOCKET,SO_REUSEADDR,(char *)&uOne, sizeof(uOne));
	}
}

BOOL CFwSocket2::Connect(LPCSTR szHost, USHORT usPort,BOOL bClearBuf,QOS *pQOS)
{
	INT iRet=1;
	SOCKADDR_IN Addr;
	INT iAddrLen=sizeof(Addr);
	WSAPROTOCOL_INFO *lpProtocolInfo=s_FwSocket2Init.GetInfo();

	if (m_bConnected) Disconnect();
	if (bClearBuf) m_InBuf.Purge();

	InitSocket(FALSE);
	WSAResetEvent(m_hEvent);
	ZeroMemory(&Addr,sizeof(Addr));

	Addr.sin_family=AF_INET;

	// check for IP address
	iRet=WSAStringToAddress((char *)szHost,AF_INET,lpProtocolInfo,(SOCKADDR *)&Addr,&iAddrLen);
	// check for DNS name
	if (iRet!=0) {
		HOSTENT *pHost=gethostbyname((const char *)szHost);
		if (!pHost) {
			Disconnect();
			WSASetLastError(WSAHOST_NOT_FOUND);
			OnError("Error gethostbyname()");
			try {
				OnConnectFin((int)WSAGetLastError());
			} catch (...) {
			}
			return FALSE;
		} else {
			// storing address information
			memcpy(&Addr.sin_addr,pHost->h_addr,pHost->h_length);
		}
	} else {
		WSAHtons(m_hSocket,usPort,&Addr.sin_port);
		WSAResetEvent(m_hEvent);
		WSAEventSelect(m_hSocket,m_hEvent,FD_CONNECT);

		iRet=WSAConnect(m_hSocket,(SOCKADDR *)&Addr,iAddrLen,NULL,NULL,pQOS,NULL);
		if (iRet!=0) {
			DWORD dwError=WSAGetLastError();
			if (dwError==WSAEWOULDBLOCK) {
				while (WSAWaitForMultipleEvents(1,&m_hEvent,FALSE,WSA_INFINITE,TRUE)==WAIT_IO_COMPLETION) ;

				// check for suceessfull connect
				timeval timeout;
				timeout.tv_sec=0;
				fd_set set;	FD_ZERO(&set); FD_SET(m_hSocket,&set);
				iRet=select(1,NULL,&set,NULL,&timeout);
				if (iRet) {
					m_bConnected=TRUE;
					WSAResetEvent(m_hEventIoThread);
					WSAResetEvent(m_hEventStopThread);
					m_hThread=(HANDLE)_beginthread((void (__cdecl *) (void *))FnIoRecvThread,2048,(LPVOID)this);
					try {
						OnConnectFin(0);
					} catch (...) {
					}
					return TRUE;
				} else {
					try {
					} catch (...) {
						Disconnect();
						WSASetLastError(WSAECONNREFUSED);
						OnConnectFin((int)WSAGetLastError());
					}
				}
			} else {
				OnError("Error WSAConnect()");
			}
		}
	}

	Disconnect();
	return FALSE;
}

BOOL CFwSocket2::Connect(const SOCKADDR* lpSockAddr,int nSockAddrLen,BOOL bClearBuf,QOS *pQOS)
{
	int iRet;
	int iAddrLen=sizeof(SOCKADDR);

	if (m_bConnected) Disconnect();
	if (bClearBuf) m_InBuf.Purge();

	InitSocket(FALSE);
	WSAResetEvent(m_hEvent);

	WSAResetEvent(m_hEvent);
	WSAEventSelect(m_hSocket,m_hEvent,FD_CONNECT);

	iRet=WSAConnect(m_hSocket,(SOCKADDR *)lpSockAddr,iAddrLen,NULL,NULL,pQOS,NULL);
	if (iRet!=0) {
	DWORD dwError=WSAGetLastError();
		if (dwError==WSAEWOULDBLOCK) {
			while (WSAWaitForMultipleEvents(1,&m_hEvent,FALSE,WSA_INFINITE,TRUE)==WAIT_IO_COMPLETION) ;

			// check for suceessfull connect
			timeval timeout;
			timeout.tv_sec=0;
			fd_set set;	FD_ZERO(&set); FD_SET(m_hSocket,&set);
			iRet=select(1,NULL,&set,NULL,&timeout);
			if (iRet) {
				m_bConnected=TRUE;
				WSAResetEvent(m_hEventIoThread);
				WSAResetEvent(m_hEventStopThread);
				m_hThread=(HANDLE)_beginthread((void (__cdecl *) (void *))FnIoRecvThread,2048,(LPVOID)this);
				try {
					OnConnectFin(0);
				} catch (...) {
				}
				return TRUE;
			} else {
				try {
					Disconnect();
					WSASetLastError(WSAECONNREFUSED);
					OnConnectFin((int)WSAGetLastError());
				} catch (...) {
				}
			}
		} else {
			OnError("Error WSAConnect()");
		}
	}
	Disconnect();
	return FALSE;
}

void CFwSocket2::ConnectAsync(LPCSTR szHost, USHORT usPort,BOOL bClearBuf,QOS *pQOS)
{
	strncpy(m_CONNECT.szHost,szHost,MAX_PATH);
	m_CONNECT.usPort=usPort;
	m_CONNECT.bClearBuf=bClearBuf;
	m_CONNECT.pQOS=pQOS;
	memset(&m_CONNECT.SockAddr,0,sizeof(m_CONNECT.SockAddr));
	m_CONNECT.iAddrLen=0;

	_beginthread((void (__cdecl *) (void *))FnIoConnectThread,2048,(LPVOID)this);
}
	
void CFwSocket2::ConnectAsync(const SOCKADDR* lpSockAddr,int nSockAddrLen,BOOL bClearBuf,QOS *pQOS)
{
	strncpy(m_CONNECT.szHost,"",MAX_PATH);
	m_CONNECT.usPort=0;
	m_CONNECT.bClearBuf=bClearBuf;
	m_CONNECT.pQOS=pQOS;
	memcpy(&m_CONNECT.SockAddr,lpSockAddr,nSockAddrLen);
	m_CONNECT.iAddrLen=nSockAddrLen;

	_beginthread((void (__cdecl *) (void *))FnIoConnectThread,2048,(LPVOID)this);
}

void CFwSocket2::FnIoConnectThread(CFwSocket2 &socket)
{
	try {
		if (socket.m_CONNECT.iAddrLen==0) {
			socket.Connect(socket.m_CONNECT.szHost,socket.m_CONNECT.usPort,socket.m_CONNECT.bClearBuf,socket.m_CONNECT.pQOS);
		} else {
			socket.Connect(&socket.m_CONNECT.SockAddr,socket.m_CONNECT.iAddrLen,socket.m_CONNECT.bClearBuf,socket.m_CONNECT.pQOS);
		}
	} catch (...) {
	}
}

void CFwSocket2::CancelOp()
{
	WSASetEvent(m_hEvent);
}

void CFwSocket2::Server(UINT uPort)
{
	int iRet;
	int iAddrLen=sizeof(SOCKADDR);

	if (m_bConnected) Disconnect();
	m_InBuf.Purge();

	InitSocket(TRUE);
	InterlockedExchange(&s_dwConcurrentConnections,0);

	SOCKADDR_IN Addr;
	Addr.sin_family=AF_INET;
	WSAHtons(m_hSocket,uPort,&Addr.sin_port);
	Addr.sin_addr.S_un.S_addr=INADDR_ANY;

	iRet=bind(m_hSocket,(SOCKADDR *)&Addr,sizeof(SOCKADDR_IN));
	if (iRet!=0) {
		OnError("Error bind()");
		return;
	}

	if (listen(m_hSocket,5)) {
		OnError("Error listen()");
		return;
	}

	WSAResetEvent(m_hEvent);
	WSAEventSelect(m_hSocket,m_hEvent,FD_ACCEPT);

	while (WSAWaitForMultipleEvents(1,&m_hEvent,FALSE,WSA_INFINITE,TRUE)==WAIT_IO_COMPLETION) ;

	INT iAddrAcceptedLen=sizeof(m_AddrAccepted);
	ZeroMemory(&m_AddrAccepted,sizeof(m_AddrAccepted));
	m_AddrAccepted.sa_family=AF_INET;
	SOCKET hClientSocket=WSAAccept(m_hSocket,&m_AddrAccepted,&iAddrAcceptedLen,NULL,NULL);
	if (hClientSocket!=INVALID_SOCKET) {
		if (GetNumConnections() < m_dwMaxConnections) {
			try {
				m_bConnected=TRUE;
				closesocket(m_hSocket);
				m_hSocket=hClientSocket;
				WSAResetEvent(m_hEventIoThread);
				WSAResetEvent(m_hEventStopThread);
				m_hThread=(HANDLE)_beginthread((void (__cdecl *) (void *))FnIoRecvThread,2048,(LPVOID)this);
					InterlockedIncrement(&s_dwConcurrentConnections);
				OnConnectFin(0);
			} catch (...) {
			}
		} else {
			try {
				Disconnect();
			} catch (...) {
			}
		}
		return;
	} else {
		try {
			Disconnect();
			WSASetLastError(WSAEFAULT);
			OnConnectFin((int)WSAGetLastError());
		} catch (...) {
		}
		return;
	}
}

void CFwSocket2::Server(UINT uPort,CRuntimeClass *pRtSock)
{
	int iRet;
	int iAddrLen=sizeof(SOCKADDR);

	if (m_bConnected) Disconnect();
	m_InBuf.Purge();

	InitSocket(TRUE);
	InterlockedExchange(&s_dwConcurrentConnections,0);

	SOCKADDR_IN Addr;
	Addr.sin_family=AF_INET;
	WSAHtons(m_hSocket,uPort,&Addr.sin_port);
	Addr.sin_addr.S_un.S_addr=INADDR_ANY;

	iRet=bind(m_hSocket,(SOCKADDR *)&Addr,sizeof(SOCKADDR_IN));
	if (iRet!=0) {
		OnError("Error bind()");
		return;
	}

	// loop to accept
	while (TRUE) {
		if (listen(m_hSocket,5)) {
			OnError("Error listen()");
			return;
		}

		WSAResetEvent(m_hEvent);
		WSAEventSelect(m_hSocket,m_hEvent,FD_ACCEPT);

		while (WSAWaitForMultipleEvents(1,&m_hEvent,FALSE,WSA_INFINITE,TRUE)==WAIT_IO_COMPLETION) ;

		int iAddrAcceptedLen=sizeof(m_AddrAccepted);
		SOCKET hClientSocket=WSAAccept(m_hSocket,&m_AddrAccepted,&iAddrAcceptedLen,NULL,NULL);
		if (hClientSocket!=INVALID_SOCKET) {
			if (GetNumConnections() < m_dwMaxConnections) {
				try {
					try {
						CFwSocket2* pSocket=(CFwSocket2 *)pRtSock->CreateObject();
						pSocket->m_bConnected=TRUE;
						pSocket->m_hSocket=hClientSocket;
						WSAResetEvent(pSocket->m_hEventIoThread);
						WSAResetEvent(pSocket->m_hEventStopThread);
						pSocket->m_bAcceptedSock=TRUE;
						pSocket->m_hThread=(HANDLE)_beginthread((void (__cdecl *) (void *))FnIoRecvThread,2048,(LPVOID)pSocket);
							InterlockedIncrement(&s_dwConcurrentConnections);
						pSocket->OnConnectFin(0);
					} catch (...) {
					}
				} catch (...) {
					break;
				}
			} else {
					closesocket(hClientSocket);
			}
		} else {
			try {
				Disconnect();
				WSASetLastError(WSAEFAULT);
			} catch (...) {
				break;
			}
			return;
		}
	}
}

void CFwSocket2::FnIoRecvThread(CFwSocket2 &socket)
{
	int iRet;
	WSABUF wbuf;
	fd_set setr;	FD_ZERO(&setr); FD_SET(socket.m_hSocket,&setr);
	fd_set sete;	FD_ZERO(&sete); FD_SET(socket.m_hSocket,&sete);

	timeval timeout;
	timeout.tv_sec=timeout.tv_usec=0;

	try {
		UINT  uInBufLen=8192;	// default buffer size
		char *pcInBuffer=(char *)realloc(NULL,uInBufLen+1);

		WSAEVENT aWSAEVENT[2];
		aWSAEVENT[0]=socket.m_hEvent;
		aWSAEVENT[1]=socket.m_hEventStopThread;

		WSAResetEvent(socket.m_hEvent);
		WSAEventSelect(socket.m_hSocket,socket.m_hEvent,FD_READ|FD_OOB|FD_CLOSE);


		for (;;) {
			// using non-blocking select call
			DWORD dwWaitRet=0;
			while ((dwWaitRet=WSAWaitForMultipleEvents(2,aWSAEVENT,FALSE,socket.m_liMaxRecvSilence,FALSE))==WAIT_IO_COMPLETION) ;

			if (dwWaitRet==WSA_WAIT_TIMEOUT) {
				if (socket.OnSilenceTimeout()) goto fin;
				else continue;
			}

			WSAResetEvent(socket.m_hEvent);
			if (!socket.m_bConnected) break;

			iRet=select(1,&setr,NULL,&sete,&timeout);
			if (FD_ISSET(socket.m_hSocket, &setr)) {
				TRACE("Can read\n");
				ULONG uCanReadCount=0,uTmp=1;
				iRet=WSAIoctl(socket.m_hSocket,FIONREAD,&uTmp,sizeof(uTmp),
					&uCanReadCount,sizeof(uCanReadCount),&uTmp,NULL,NULL);
				if (uCanReadCount==0) {
					if (socket.m_hShellProcess!=INVALID_HANDLE_VALUE) {
						TerminateProcess(socket.m_hShellProcess,0);
						socket.m_hShellProcess=INVALID_HANDLE_VALUE;
						CloseHandle(socket.m_hShellProcess);
					}
					try {
						socket.m_bConnected=FALSE;
						socket.OnDisconnect(0); 
					} catch (...) {
					}
					break;
				} else {
					DWORD wdFlags=0;
					if (uInBufLen<uCanReadCount) {
						uInBufLen=uCanReadCount;
						pcInBuffer=(char *)realloc(NULL,uInBufLen+1);
					}
					wbuf.buf=pcInBuffer;
					wbuf.len=uInBufLen;
	retry:
					iRet=WSARecv(socket.m_hSocket,&wbuf,1,&uTmp,&wdFlags,NULL,NULL);
					if (iRet==0) {
						socket.m_Stats.m_bytes_in+=uTmp;
						socket.InBufferLock();
						pcInBuffer[uTmp]='\x0';
						socket.OnDataReceive(wbuf.buf,uTmp);
						socket.InBufferAdd(wbuf.buf,uTmp);
						try {
							socket.OnDataReceiveFin();
						} catch (...) {
						}
						socket.InBufferUnlock();
						WSASetEvent(socket.m_hEventDataReady);

						if (wbuf.len<uCanReadCount) {
							wbuf.buf+=uTmp;
							wbuf.len-=uTmp;
							goto retry;
						}
					} else {
						socket.OnError("WSARecv() error");
					}
				}
			}
			if (FD_ISSET(socket.m_hSocket, &sete)) {
				TRACE("OOB is present\n");
			}
		}
fin:
		socket.m_hThread=INVALID_HANDLE_VALUE;
		socket.m_bConnected=FALSE;
		free(pcInBuffer);

		WSASetEvent(socket.m_hEventIoThread);

		// deleting object after termonation of connection
		if (socket.m_bAcceptedSock) {
			socket.m_bAcceptedSock=FALSE;
			closesocket(socket.m_hSocket);
			socket.m_hSocket=INVALID_SOCKET;
			InterlockedDecrement(&s_dwConcurrentConnections);
			delete &socket;
		}
	} catch (...) {
		TRACE("CFwSocket2::FnIoRecvThread() Abnorman end.");
	}


	_endthread();
}

void CFwSocket2::OnConnectFin(int iErrorCode)
{
	if (!iErrorCode) {
		TRACE("Connected\n");
	} else {
		TRACE("Not connected\n");
	}
}

void CFwSocket2::OnDisconnect(int iErrorCode)
{
	TRACE("Disconnected\n");
}

void CFwSocket2::OnDataReceive(LPCSTR szBuffer,DWORD &uLen)
{
}

void CFwSocket2::OnDataReceiveFin()
{
}

void CFwSocket2::OnError(LPCSTR pszComment)
{
	TRACE("\n%s\n%s\n",pszComment,GetErrorString(WSAGetLastError()));
}

BOOL CFwSocket2::OnSilenceTimeout()
{
	TRACE("OnSilenceTimeout() Maximum silence timeout reached\n");
	return FALSE;
}

void CFwSocket2::Disconnect()
{
	try {
		CancelOp();
		if (IsConnected()) {
			if (m_hShellProcess!=INVALID_HANDLE_VALUE) {
				TerminateProcess(m_hShellProcess,0);
				m_hShellProcess=INVALID_HANDLE_VALUE;
				CloseHandle(m_hShellProcess);
			}
			m_bConnected=FALSE;
			WSASetEvent(m_hEventStopThread);
			WSAWaitForMultipleEvents(1,&m_hEventIoThread,FALSE,WSA_INFINITE,TRUE);
			closesocket(m_hSocket);
		}
		if (m_hSocket!=INVALID_SOCKET) {
			closesocket(m_hSocket);
			m_hSocket=INVALID_SOCKET;
		}
	} catch (...) {
	}
}

CString CFwSocket2::GetErrorString(int iErrno,BOOL bDesc)
{
	typedef struct tag_error_struct {
	    int		errnum;
		LPCTSTR	errstr;
	    LPCTSTR	errstrdesc;
	} ERROR_STRUCT;

	const ERROR_STRUCT errlist[] = {
		{WSAEINTR,           _T("WSAEINTR"),_T("Interrupted")},
		{WSAEBADF,           _T("WSAEBADF"),_T("Bad file number")},
		{WSAEFAULT,          _T("WSAEFAULT"),_T("Bad address")},
		{WSAEINVAL,          _T("WSAEINVAL"),_T("Invalid argument")},
		{WSAEMFILE,          _T("WSAEMFILE"),_T("Too many open files")},
		// Windows Sockets definitions of regular Berkeley error constants
		{WSAEWOULDBLOCK,     _T("WSAEWOULDBLOCK"),_T("Socket marked as non-blocking")},
		{WSAEINPROGRESS,     _T("WSAEINPROGRESS"),_T("Blocking call in progress")},
		{WSAEALREADY,        _T("WSAEALREADY"),_T("Command already completed")},
		{WSAENOTSOCK,        _T("WSAENOTSOCK"),_T("Descriptor is not a socket")},
		{WSAEDESTADDRREQ,    _T("WSAEDESTADDRREQ"),_T("Destination address required")},
		{WSAEMSGSIZE,        _T("WSAEMSGSIZE"),_T("Data size too large")},
		{WSAEPROTOTYPE,      _T("WSAEPROTOTYPE"),_T("Protocol is of wrong type for this socket")},
		{WSAENOPROTOOPT,     _T("WSAENOPROTOOPT"),_T("Protocol option not supported for this socket type")},
		{WSAEPROTONOSUPPORT, _T("WSAEPROTONOSUPPORT"),_T("Protocol is not supported")},
		{WSAESOCKTNOSUPPORT, _T("WSAESOCKTNOSUPPORT"),_T("Socket type not supported by this address family")},
		{WSAEOPNOTSUPP,      _T("WSAEOPNOTSUPP"),_T("Option not supported")},
		{WSAEPFNOSUPPORT,    _T("WSAEPFNOSUPPORT"),_T("")},
		{WSAEAFNOSUPPORT,    _T("WSAEAFNOSUPPORT"),_T("Address family not supported by this protocol")},
		{WSAEADDRINUSE,      _T("WSAEADDRINUSE"),_T("Address is in use")},
		{WSAEADDRNOTAVAIL,   _T("WSAEADDRNOTAVAIL"),_T("Address not available from local machine")},
		{WSAENETDOWN,        _T("WSAENETDOWN"),_T("Network subsystem is down")},
		{WSAENETUNREACH,     _T("WSAENETUNREACH"),_T("Network cannot be reached")},
		{WSAENETRESET,       _T("WSAENETRESET"),_T("Connection has been dropped")},
		{WSAECONNABORTED,    _T("WSAECONNABORTED"),_T("Connection aborted")},
		{WSAECONNRESET,      _T("WSAECONNRESET"),_T("Connection resed by peer")},
		{WSAENOBUFS,         _T("WSAENOBUFS"),_T("No buffer space available")},
		{WSAEISCONN,         _T("WSAEISCONN"),_T("Socket is already connected")},
		{WSAENOTCONN,        _T("WSAENOTCONN"),_T("Socket is not connected")},
		{WSAESHUTDOWN,       _T("WSAESHUTDOWN"),_T("Socket has been shut down")},
		{WSAETOOMANYREFS,    _T("WSAETOOMANYREFS"),_T("Too many references to some kernel object")},
		{WSAETIMEDOUT,       _T("WSAETIMEDOUT"),_T("Command timed out")},
		{WSAECONNREFUSED,    _T("WSAECONNREFUSED"),_T("Connection refused")},
		{WSAELOOP,           _T("WSAELOOP"),_T("Cannot translate name")},
		{WSAENAMETOOLONG,    _T("WSAENAMETOOLONG"),_T("Name component or name was too long")},
		{WSAEHOSTDOWN,       _T("WSAEHOSTDOWN"),_T("A socket operation failed because the destination host was down")},
		{WSAEHOSTUNREACH,    _T("WSAEHOSTUNREACH"),_T("A socket operation was attempted to an unreachable host")},
		{WSAENOTEMPTY,       _T("WSAENOTEMPTY"),_T("Cannot remove a directory that is not empty")},
		{WSAEPROCLIM,        _T("WSAEPROCLIM"),_T("A Windows Sockets implementation may have a limit on the number of applications that may use it simultaneously")},
		{WSAEUSERS,          _T("WSAEUSERS"),_T("Ran out of quota")},
		{WSAEDQUOT,          _T("WSAEDQUOT"),_T("Ran out of disk quota")},
		{WSAESTALE,          _T("WSAESTALE"),_T("File handle reference is no longer available")},
		{WSAEREMOTE,         _T("WSAEREMOTE"),_T("Item is not available locally")},
		// EXTENDED Windows Sockets error constant definitions
		{WSASYSNOTREADY,     _T("WSASYSNOTREADY"),_T("Network subsystem not ready")},
		{WSAVERNOTSUPPORTED, _T("WSAVERNOTSUPPORTED"),_T("Version not supported")},
		{WSANOTINITIALISED,  _T("WSANOTINITIALISED"),_T("WSAStartup() has not been successfully called")},
		// Other error constants.
		{WSAHOST_NOT_FOUND,  _T("WSAHOST_NOT_FOUND"),_T("Host not found")},
		{WSATRY_AGAIN,       _T("WSATRY_AGAIN"),_T("Host not found or SERVERFAIL")},
		{WSANO_RECOVERY,     _T("WSANO_RECOVERY"),_T("Non-recoverable error")},
		{WSANO_DATA,         _T("WSANO_DATA"),_T("No data record of requested type")},
		{-1,                 NULL,NULL}
	};

	for (int i=0; ; i++) {
		if (errlist[i].errnum==-1) break;
		if (iErrno==errlist[i].errnum) {
			CString strRet;
			if (bDesc) {
				strRet+=errlist[i].errstr;
				strRet+=_T(" - ");
			}
			return strRet+errlist[i].errstrdesc;
		}
	}

	return _T("No error");
}

BOOL CFwSocket2::InBufferAdd(char ch)
{
	return m_InBuf.Add(ch);
}

BOOL CFwSocket2::InBufferAdd(LPCSTR szStr)
{
	return m_InBuf.Add(szStr);
}

BOOL CFwSocket2::InBufferAdd(LPCVOID pvMem,UINT uLen)
{
	return m_InBuf.Add(pvMem,uLen);
}

BOOL CFwSocket2::InBufferDel(UINT uLen,UINT uBegin)
{
	return m_InBuf.Del(uLen,uBegin);
}

void CFwSocket2::InBufferLock()
{
	m_InBuf.Lock();
}

void CFwSocket2::InBufferUnlock()
{
	m_InBuf.Unlock();
}


CFwSocket2::CTcpIpStatistics& CFwSocket2::GetStatistics()
{
	return m_Stats;
}

CFwSocket2::CBuffer& CFwSocket2::GetInBuffer() 
{
	return m_InBuf;
}

BOOL CFwSocket2::Get(void *pvBuffer,UINT uLength)
{
	UINT bRetv=FALSE;
	m_InBuf.Lock();
	if (pvBuffer && uLength<=m_InBuf.Len()) {
		memcpy(pvBuffer,m_InBuf.Buf(),uLength);
		m_InBuf.Del(uLength);
		bRetv=TRUE;
	}
	m_InBuf.Unlock();
	return bRetv;
}

BOOL CFwSocket2::Put(LPCVOID pvBuffer,UINT uLength)
{
	UINT retv=FALSE;
	EnterCriticalSection(&m_CrsPut);
	if (pvBuffer) {
		int iRet=0;
		DWORD dwBytesSent=0;
		WSABUF wbuf;
retry:
		dwBytesSent=0;
		wbuf.buf=(char *)pvBuffer;
		wbuf.len=uLength;
		iRet=WSASend(m_hSocket,&wbuf,1,&dwBytesSent,0,NULL,NULL);
		if (iRet==0) {
			if (dwBytesSent<uLength) {
				pvBuffer=((BYTE *)pvBuffer)+dwBytesSent;
				uLength-=dwBytesSent;
				goto retry;
			}
		} else {
			OnError("WSASend() error");
		}
		retv=TRUE;
	}
	LeaveCriticalSection(&m_CrsPut);
	return retv;
}

BOOL CFwSocket2::Gets(CString& str)
{
	BOOL retv=FALSE;
	UINT uLen=0;

	str="";
	m_InBuf.Lock();
	char *pBuf=(char *)m_InBuf.Buf();
	char *pNl=NULL;
	pNl=strchr(pBuf,'\r');
	if (!pNl) pNl=strchr(pBuf,'\n');
	if (!pNl && !IsConnected()) pNl=pBuf+m_InBuf.Len();
	if (pNl) {
		uLen=pNl-pBuf;
		for (UINT i=0; i<uLen; i++) {
			char c=pBuf[i];
			switch (c) {
			case '\r': break;
			case '\n': break;
			case '\b': 
				{
					int iStrLn=str.GetLength();
					if (iStrLn>0) str=str.Left(iStrLn-1);
					break;
				}
			default: str+=pBuf[i]; 
			}
		}
		if (uLen>0) retv=TRUE;
	}
	m_InBuf.Unlock();
	m_InBuf.Del(uLen);
	if (*pBuf=='\r') m_InBuf.Del(1);
	if (*pBuf=='\n') m_InBuf.Del(1);
	return retv;
}

BOOL CFwSocket2::Puts(CString str,BOOL bNl)
{
	if (bNl) str+="\n";
	return Put((void *)(LPCSTR)str,str.GetLength());
}

BOOL CFwSocket2::Putf(LPCSTR fmt,...)
{
	CString tmp;
	va_list pvar;

	va_start(pvar,fmt);
	tmp.FormatV(fmt,pvar);
	va_end(pvar);
	
	return Put((void *)(LPCSTR)tmp,tmp.GetLength());
}


void CFwSocket2::ClearIOBuffers()
{
	m_InBuf.Purge();
}


BOOL CFwSocket2::ExecuteCommand(LPCSTR szCmdLine)
{
	BOOL bRetv=TRUE;

	HANDLE hWrStdInput=NULL;
	HANDLE hRdStdOutput=NULL;
	HANDLE hRdStdError=NULL;

	STARTUPINFO si;
	memset(&si,0,sizeof(STARTUPINFO));
	si.cb=sizeof(STARTUPINFO);
	si.dwFlags=STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
	si.wShowWindow=SW_HIDE;

	// creating I/O pipes
	SECURITY_ATTRIBUTES saAttr;
	saAttr.nLength=sizeof(SECURITY_ATTRIBUTES);
	saAttr.bInheritHandle=TRUE;
	saAttr.lpSecurityDescriptor=NULL; 

	if (!CreatePipe(&si.hStdInput,&hWrStdInput,&saAttr,0)) goto error;
	SetStdHandle(STD_INPUT_HANDLE,si.hStdInput);
	if (!CreatePipe(&hRdStdOutput,&si.hStdOutput,&saAttr,0)) goto error;
	SetStdHandle(STD_OUTPUT_HANDLE,si.hStdOutput);
	if (!CreatePipe(&hRdStdError,&si.hStdError,&saAttr,0)) goto error;
	SetStdHandle(STD_ERROR_HANDLE,si.hStdError);


	PROCESS_INFORMATION pi;

	bRetv=CreateProcess(
		NULL,  // name of executable module
		(char *)(LPCSTR)szCmdLine,	// command line string
		NULL, 
		NULL, 
		TRUE,						// handle inheritance flag
		NULL,						// creation flags
		NULL,						// new environment block
		NULL,						// current directory name
		&si, 
		&pi
		);

	if (bRetv) {
		m_hShellProcess=pi.hProcess;

		// closing the rest of I/O pipes
		CloseHandle(si.hStdInput);
		CloseHandle(si.hStdOutput);
		CloseHandle(si.hStdError);

		// waiting for process termonation
		DWORD dwExCode=STILL_ACTIVE;
		SetPriorityClass(pi.hProcess,NORMAL_PRIORITY_CLASS);

		while (dwExCode==STILL_ACTIVE) {
			DWORD dw;
			DWORD dwReady=0;

			if (!GetExitCodeProcess(pi.hProcess,&dwExCode)) {
				m_hShellProcess=INVALID_HANDLE_VALUE;
				break;
			}

			// check for TELNET user is connected
			if (!IsConnected()) {
				break;
			}

			// processing TELNET user input
			if (m_InBuf.m_uBufLen>0) {
				DWORD dwBytes=0;
				BYTE *pBufToWrite;
				dwBytes=m_InBuf.m_uBufLen;
				pBufToWrite=m_InBuf.m_pBuf;
				m_InBuf.Lock();
retry:
				if (!WriteFile(hWrStdInput,m_InBuf.m_pBuf,m_InBuf.m_uBufLen,&dw,NULL)) {
					m_InBuf.m_uBufLen=0;
					m_InBuf.Unlock();
					break;
				} else {
					if (dw<dwBytes) {
						dwBytes-=dw;
						pBufToWrite+=dw;
						goto retry;
					}
				}
				m_InBuf.m_uBufLen=0;
				m_InBuf.Unlock();
			}

			// 

			// processing shell output
			if (PeekNamedPipe(hRdStdOutput,NULL,0,NULL,&dwReady,NULL)) {
				if (dwReady) {
					char *szTmpBuf=new char[dwReady+1];
					if (ReadFile(hRdStdOutput,szTmpBuf,dwReady,&dw,NULL))
						szTmpBuf[dw]=0;
						char *pBuf=szTmpBuf;
						Put(pBuf,dw-(pBuf-szTmpBuf));
					delete[] szTmpBuf;
				}
			} else {
				CloseHandle(m_hShellProcess);
				m_hShellProcess=INVALID_HANDLE_VALUE;
				break;
			}

			// processing shell stderr
			if (PeekNamedPipe(hRdStdError,NULL,0,NULL,&dwReady,NULL)) {
				if (dwReady) {
					char *szTmpBuf=new char[dwReady+1];
					if (ReadFile(hRdStdError,szTmpBuf,dwReady,&dw,NULL))
						szTmpBuf[dw]=0;
						char *pBuf=szTmpBuf;
						Put(pBuf,dw-(pBuf-szTmpBuf));
					delete[] szTmpBuf;
				}
			} else {
				CloseHandle(m_hShellProcess);
				m_hShellProcess=INVALID_HANDLE_VALUE;
				break;
			}
			Sleep(1);
		}
		
		// closing the rest of I/O pipes
		CloseHandle(hWrStdInput);
		CloseHandle(hRdStdOutput);
		CloseHandle(hRdStdError);

		if (m_hShellProcess!=INVALID_HANDLE_VALUE) {
			CloseHandle(m_hShellProcess);
			m_hShellProcess=INVALID_HANDLE_VALUE;
		}

		// clear buffer
		m_InBuf.Lock();
		m_InBuf.m_uBufLen=0;
		m_InBuf.Unlock();
	} else {
error:
		bRetv=FALSE;
	}

	return bRetv;
}

void CFwSocket2::WaitDataEvent()
{
	WSAEVENT ev[2];
	ev[0]=m_hEventIoThread;
	ev[1]=m_hEvent;
	WSAWaitForMultipleEvents(2,ev,FALSE,WSA_INFINITE,TRUE);
	WSAResetEvent(m_hEventIoThread);
}
void CFwSocket2::SetMaxConnections(LONG dwNum)
{
	m_dwMaxConnections=dwNum;
}

LONG CFwSocket2::GetNumConnections()
{
	LONG dwRet=0;
	InterlockedExchange(&dwRet,s_dwConcurrentConnections);
	return dwRet;
}

SOCKADDR * CFwSocket2::GetAcceptedAddress()
{
	return &m_AddrAccepted;
}

void CFwSocket2::SetMaxRecvSilence(LONG iTimeout)
{
	m_liMaxRecvSilence=iTimeout;
}
