/*!
	\file
	FwSocket2.h : header file.
	$Author: Arcady Pchelintsev $
	$Date: 2001/10/19 11:16:54  $
	$State: Exp $
	$Source: \\PRAGMA\DATAC\\FwLib\\FwLib32\FwSocket2.h $
	$Revision:$	1.0.0.0
\n  $Log: $
*/


#if !defined(AFX_FWSOCKET2_H__9DDFE080_E2D9_4B09_B746_17D3373DB1AB__INCLUDED_)
#define AFX_FWSOCKET2_H__9DDFE080_E2D9_4B09_B746_17D3373DB1AB__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <winsock2.h>
#include <Ws2tcpip.h>

/*! \ingroup Io
	\class CFwSocket2
    \brief This class is thread-safe winsock2 wrapper.
*/
class CFwSocket2 : public CObject
{
	DECLARE_DYNCREATE(CFwSocket2)

	// Internal input and output queue
	class CBuffer {
		friend class CFwSocket2;
		CRITICAL_SECTION m_Crs;
		BYTE *m_pBuf;
		volatile UINT m_uBufLen;
		volatile UINT m_uBufLenMax;
		CFwSocket2 *m_pIp;
		// data manipulation access functions
		BOOL Add(char ch);					// add single character
		BOOL Add(LPCSTR szStr);				// add ASCIIZ string
		BOOL Add(LPCVOID pvMem,UINT uLen);	// add binary data
		BOOL Del(UINT uNumber,UINT uBegin=0);		// delete and drag tail
		void Purge();
		// sync functions
		inline void Lock() {EnterCriticalSection(&m_Crs);}
		inline void Unlock() {LeaveCriticalSection(&m_Crs);}
	public:
		CBuffer();
		~CBuffer();
		// member functions
		const BYTE * Buf() {return m_pBuf;}	// buffer accecc
		UINT Len() {return m_uBufLen;}		// data length
		BOOL Realloc(UINT uSize);			// buffer realloc
		// operators
		BYTE & operator[](int i) {return m_pBuf[i];}
	} m_InBuf;

	// statistics
	struct CTcpIpStatistics {
		volatile DWORD m_bytes_in;		//!< bytes received by object since last connect
		volatile DWORD m_bytes_out;		//!< bytes sent by object since last connect
	} m_Stats;

	friend class CFwSocket2::CBuffer;
public:
	CFwSocket2();						//!< default conctructor
	virtual ~CFwSocket2();				//!< default destructor

	//! Connects to remote server using host name or IP (blocking)
	BOOL Connect(LPCSTR szHost,USHORT usPort,BOOL bClearBuf=TRUE,QOS *pQOS=NULL);
	//! Connects to remote server using SOCKADDR structure (non-blocking)
	void ConnectAsync(LPCSTR szHost,USHORT usPort,BOOL bClearBuf=TRUE,QOS *pQOS=NULL);
	//! Connects to remote server using SOCKADDR structure (blocking)
	BOOL Connect(const SOCKADDR* lpSockAddr,int nSockAddrLen,BOOL bClearBuf=TRUE,QOS *pQOS=NULL);
	//! Connects to remote server using SOCKADDR structure (non-blocking)
	void ConnectAsync(const SOCKADDR* lpSockAddr,int nSockAddrLen,BOOL bClearBuf=TRUE,QOS *pQOS=NULL);
	//! Bind to specified port and become IP-server (blocking), easy version
	void Server(UINT uPort);
	//! Bind to specified port and become IP-server (blocking), professional version
	void Server(UINT uPort,CRuntimeClass *pRtSock);
	//! Cancel Connect() or Server() blocking operations
	void CancelOp();
	//! returns TRUE if connected
	BOOL IsConnected() {return m_bConnected;}
	//! Gets memory block from opened socket (non-blocking)
	static CString GetErrorString(int iErrno,BOOL bDesc=FALSE);
	//! Gets memory block from opened socket (non-blocking)
	BOOL Get(void *pvBuffer,UINT uLength);
	//! Gets string from opened socket (non-blocking)
	BOOL Gets(CString& str);
	//! Wait for incoming data
	void WaitDataEvent();

	//! Puts string to opened socket (non-blocking)
	BOOL Put(LPCVOID pvBuffer,UINT uLength);
	//! Puts string to opened socket (non-blocking)
	BOOL Puts(CString str,BOOL bNl=TRUE);
	//! Puts formatted string to opened socket (non-blocking)
	BOOL Putf(LPCSTR szFmt,...);
	//! Disconnect from remote server (blocking)
	void Disconnect();
	//! Clear all buffers
	void ClearIOBuffers();
	//! Execute shell command (for server only)
	BOOL ExecuteCommand(LPCSTR szCmdLine);

// Buffers access and statistics access
	CTcpIpStatistics& GetStatistics();			//!< get TCP/IP statistics for session
	CBuffer& GetInBuffer();						//!< Gets access to input buffer
	// input buffer
	BOOL InBufferAdd(char ch);					//!< Add byte to input buffer
	BOOL InBufferAdd(LPCSTR szStr);				//!< Add string to input buffer
	BOOL InBufferAdd(LPCVOID pvMem,UINT uLen);	//!< Add memory block to input buffer
	BOOL InBufferDel(UINT uLen,UINT uBegin=0);	//!< Remove block from input buffer
	void InBufferLock();						//!< Lock input buffer
	void InBufferUnlock();						//!< Unlock input buffer

	SOCKADDR * GetAcceptedAddress();			//!< Returns accepted address
	LONG GetNumConnections();					//!< get number of connections
	void SetMaxConnections(LONG dwNum);			//!< set connection limit for server mode

	//! set max timeout for silence (can be used only in constuctor of derived class)
	void SetMaxRecvSilence(LONG iTimeout=WSA_INFINITE);		


protected:
// Overridable Notification Functions

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CFwSocket)
	virtual void OnDisconnect(int iErrorCode);	//!< Disconnect event handler
	virtual void OnConnectFin(int nErrorCode);	//!< Connect ready (or error) event handler
	virtual void OnDataReceive(LPCSTR szBuffer,DWORD &uLen);	//!< Data ready but not yet in input buffer event handler
	virtual void OnDataReceiveFin();			//!< Data ready event handler
	virtual BOOL OnSilenceTimeout();			//!< Maximum silence timeout reached (return 1 to disconnect)
	virtual void OnError(LPCSTR pszComment);	//!< Error event handler
	//}}AFX_VIRTUAL


protected:
	// internal threads
	static void FnIoRecvThread(CFwSocket2 &socket);		// I/O and messaging thread 
	static void FnIoConnectThread(CFwSocket2 &socket);	// connecting thread

	// internal member functions
	void InitSocket(BOOL bServerSocket);	// socket creation

protected:
	CRITICAL_SECTION m_CrsPut;
	SOCKADDR m_AddrAccepted;
	INT m_dwMaxConnections;					// max number of allowed connections
	LONG m_liMaxRecvSilence;
	volatile BOOL m_bConnected;				// boolean flag, connected if TRUE
	volatile BOOL m_bAcceptedSock;			// boolean flag, connected if accepted
	volatile HANDLE m_hThread;				// main I/O and messaging thread handle
	volatile HANDLE m_hShellProcess;		// shell process handle for server
	WSAOVERLAPPED m_WSAOVERLAPPED;			// for overlapped operations
	volatile SOCKET m_hSocket;						// socket handle
	WSAEVENT m_hEvent;						// event for all operations
	WSAEVENT m_hEventStopThread;			// event for all operations
	WSAEVENT m_hEventIoThread;				// event thread termination
	WSAEVENT m_hEventDataReady;				// event for blocking reads
	// for async connect
	struct {
		CHAR    szHost[MAX_PATH+1];
		USHORT  usPort;
		BOOL	bClearBuf;
		QOS		*pQOS;
		SOCKADDR SockAddr;
		int iAddrLen;
	} m_CONNECT;
};

#endif // !defined(AFX_FWSOCKET2_H__9DDFE080_E2D9_4B09_B746_17D3373DB1AB__INCLUDED_)
